from pathlib import Path
from pydotfiles.loading import load_plist
from pydotfiles.utils import run_command


class DockManager:
    """
    Manages anything related to a dock
    """

    @property
    def dock_plist_path(self):
        return Path.home().joinpath("Library/Preferences/com.apple.dock.plist")

    def get_current_dock_applications(self):
        dock_plist_data = load_plist(self.dock_plist_path)
        return extract_persistent_app_names(dock_plist_data)

    @staticmethod
    def delete_dock_plist_file():
        command = f'defaults delete com.apple.Dock persistent-apps'
        run_command(command)

    @staticmethod
    def add_application_to_default_on_dock(application):
        command = f'defaults write com.apple.dock persistent-apps -array-add "<dict><key>tile-data</key><dict><key>file-data</key><dict><key>_CFURLString</key><string>/Applications/{application}.app</string><key>_CFURLStringType</key><integer>0</integer></dict></dict></dict>"'
        run_command(command)

    @staticmethod
    def restart_dock():
        command = f"killall Dock"
        run_command(command)


"""
Helper methods
"""


def extract_persistent_app_names(plist_data):
    persistent_apps = plist_data['persistent-apps']
    return {persistent_app['tile-data']['file-label'] for persistent_app in persistent_apps}



# General imports
from dataclasses import dataclass
from typing import Dict


@dataclass(frozen=True, order=True)
class Profile:
    name: str
    dependencies: list[str]

    @staticmethod
    def from_dict(data: Dict) -> list['Profile']:
        return [Profile.from_profile_dict(name, dependencies) for name, dependencies in data.items()]

    @staticmethod
    def from_profile_dict(name: str, dependencies: list[str]) -> 'Profile':
        return Profile(
            name=name,
            dependencies=dependencies
        )



from enum import Enum, auto
from pydotfiles.utils import BijectiveDictionary


"""
Model enums
"""


class FileActionType(Enum):
    """
    Represents what kind of file operation
    this FileAction is
    """
    COPY = auto()
    DELETE = auto()

    SYMLINK = auto()
    UNSYMLINK = auto()

    SCRIPT = auto()
    UNDO_SCRIPT = auto()

    @staticmethod
    def from_string(label):
        if label is None:
            raise KeyError("Enum: No file action was passed in")
        return FileActionType[label.upper()]

    @staticmethod
    def get_reverse(file_action_type):
        file_action_reverse_mapping = BijectiveDictionary({
            FileActionType.COPY: FileActionType.DELETE,
            FileActionType.SYMLINK: FileActionType.UNSYMLINK,
            FileActionType.SCRIPT: FileActionType.UNDO_SCRIPT
        })
        return file_action_reverse_mapping.get(file_action_type)


class OverrideAction(Enum):
    """
    Represents an override action that
    a user can input when deciding what
    to do about certain file actions
    """

    SKIP_FILE = auto()
    SKIP_ALL_FILES = auto()

    OVERWRITE_FILE = auto()
    OVERWRITE_ALL_FILES = auto()

    BACKUP_FILE = auto()
    BACKUP_ALL_FILES = auto()

    @staticmethod
    def from_label(label):
        if label is None:
            raise KeyError("Enum: No override action was passed in")
        return OverrideAction[label.upper()]

    @staticmethod
    def affects_multiple_files(override_action):
        return override_action in {
            OverrideAction.SKIP_ALL_FILES,
            OverrideAction.OVERWRITE_ALL_FILES,
            OverrideAction.BACKUP_ALL_FILES
        }


"""
Exception enums
"""


class PydotfilesErrorReason(Enum):
    """
    Represents a particular reason why
    the pydotfiles failed, enables extraction
    of a potentially useful help message
    as well
    """

    UNKNOWN_ERROR = auto()
    NO_REMOTE_REPO = auto()
    REMOTE_REPO_CLONE_ISSUE = auto()
    UNKNOWN_CLEANING_TARGET = auto()
    UNKNOWN_MODULE_NAME = auto()

    @staticmethod
    def get_help_message(reason):
        help_message_map = {
            PydotfilesErrorReason.UNKNOWN_ERROR: None
        }
        return help_message_map.get(reason)


class ValidationErrorReason(Enum):
    """
    Represents a particular reason why
    the validation of a given dotfiles,
    module, or file failed
    """

    INVALID_TARGET = auto()

    INVALID_EMPTY_FILE = auto()
    INVALID_SYNTAX = auto()
    INVALID_SCHEMA = auto()

    # Specific schema issues
    INVALID_DATA = auto()
    INVALID_SCHEMA_VERSION = auto()
    INVALID_SCHEMA_TYPE = auto()

    @staticmethod
    def get_help_message(reason):
        help_message_map = {
            ValidationErrorReason.INVALID_EMPTY_FILE: "An empty invalid configuration file was detected",
            ValidationErrorReason.INVALID_SCHEMA: "A given configuration file contains an invalid schema"
        }
        return help_message_map.get(reason)



# General imports
from os import walk as os_walk
from os import path as os_path
from pathlib import Path
from typing import Dict
from json import load as json_load
from logging import getLogger

# 3rd party imports
from yaml import load as yaml_load
from yaml import YAMLError, Loader
from yaml.scanner import ScannerError

# Project imports
from pydotfiles.v4.common import Configuration
from pydotfiles.v4.common import PydotfilesError
from pydotfiles.v4.validator import validate

logger = getLogger(__name__)

##
# Publicly accessible methods
##


def load_configuration(directory: Path) -> list[Configuration]:
    # Generates a set of files that we need to validate from a tree structure
    potential_configuration_files = __load_potential_configuration_files(directory)
    loaded_potential_configs = {config_file: __load_data_from_file(config_file) for config_file in potential_configuration_files}
    return validate(loaded_potential_configs)

##
# Helper methods
##


def __load_potential_configuration_files(directory: Path) -> set[Path]:
    potential_configuration_files = set()

    exclude = {".git"}
    for path_prefix, directory_names, file_names in os_walk(directory, topdown=True):
        # We want to exclude certain directories from our loading walk. For more
        # information, see https://stackoverflow.com/a/19859907
        directory_names[:] = [d for d in directory_names if d not in exclude]
        for file_name_path in file_names:
            if file_name_path.endswith(".json") or file_name_path.endswith(".yaml") or file_name_path.endswith(".yml"):
                potential_configuration_files.add(Path(os_path.join(path_prefix, file_name_path)))

    # for file in initial_files_to_validate:
    #     print(f"files to potentially load from: {file}")
    return potential_configuration_files


def __load_data_from_file(config_file: Path) -> Dict:
    """
    Loads a configuration file
    """
    try:
        with config_file.open('r') as config_fd:
            if str(config_file).endswith("json"):
                return json_load(config_fd)
            elif str(config_file).endswith("yaml") or str(config_file).endswith("yml"):
                try:
                    return yaml_load(config_fd, Loader)
                except (YAMLError, ScannerError) as e:
                    load_error = PydotfilesError.due_to_yaml_error()
                    load_error.context_map['file'] = str(config_file)
                    load_error.context_map['error'] = e
                    raise load_error from e
            else:
                load_error = PydotfilesError.due_to_unsupported_config_filetype()
                load_error.context_map['file'] = str(config_file)
                raise load_error
    except PydotfilesError as e:
        logger.error(e.help_message)



# General imports
from dataclasses import dataclass
from enum import Enum
from typing import Dict


class Action(Enum):
    copy = "copy"
    symlink = "symlink"
    script = "script"


@dataclass(frozen=True, order=True)
class ActionType:
    action: Action
    files: Dict[str, str]
    hidden: bool
    sudo: bool
    absolute: bool

    @staticmethod
    def from_dict(data: list[Dict]) -> list['ActionType']:
        return [ActionType.from_action_dict(datum) for datum in data]

    @staticmethod
    def from_action_dict(data: Dict) -> 'ActionType':
        return ActionType(
            action=Action[data.get("action")],
            files=data.get("files"),
            hidden=False if data.get("hidden") is None else data.get("hidden"),
            sudo=False if data.get("sudo") is None else data.get("sudo"),
            absolute=False if data.get("absolute") is None else data.get("absolute"),
        )

    def to_dict(self):
        return {
            'action': self.action.value,
            'files': self.files,
            'hidden': self.hidden,
            'sudo': self.sudo,
            'absolute': self.absolute,
        }



# General imports
from pathlib import Path
from typing import Dict, Optional, Set
from collections import defaultdict
from logging import getLogger
from tempfile import mkdtemp as mktempdir
from datetime import datetime
from os import walk as os_walk
from os import path as os_path
from dataclasses import dataclass
from shutil import copy2 as shutil_copy2
from shutil import rmtree as shutil_rmtree
from os import chmod as os_chmod
from zipfile import ZipFile

# Project imports
from pydotfiles.v4.common import Configuration
from pydotfiles.v4.common.alpha import AlphaManifest, AlphaCore, AlphaDefaultSettings, AlphaDeveloperEnvironments, \
    OSName, ActionType, OperatingSystem, Environment
from pydotfiles.v4.common.alpha.action import Action
from .io import template_run_script, template_copy_file, template_symlink_file, template_starter_helper_functions, template_dev_environment_setup
from .os import template_os_packages, template_os_applications, template_os_default_dock, template_os_default_setting_files

logger = getLogger(__name__)


class Builder:
    """
    Generates profile build packages, which are self-contained,
    'single use' dotfiles that aren't dynamically managed by pydotfiles,
    but are instead meant for easy 1-time installations.

    This is to help support a lighter-weight entry into automated dotfiles,
    without requiring users to commit to a whole dynamically updated ecosystem
    """

    def __init__(self, configurations: list[Configuration], base_dir: Path, output_path: Path, is_release: bool):
        # Maps from a given profile name to the file paths that are defined there
        self.profile_map: Dict[str, list[Path]] = defaultdict(list)
        self.profile_dependencies: Dict[str, list[str]] = defaultdict(list)
        self.profile_file_types: Dict[Path, type] = {}
        self.dev_environments: Dict[Path, AlphaDeveloperEnvironments] = {}
        self.default_settings: Dict[Path, AlphaDefaultSettings] = {}
        self.core: Dict[Path, AlphaCore] = {}
        self.defined_os: set[OSName] = set()
        self.base_dir = base_dir
        self.output_path = output_path
        self.is_release = is_release

        for configuration in configurations:
            if isinstance(configuration.data, AlphaCore):
                self.core[configuration.file_path] = configuration.data
                self.profile_map[configuration.data.profile].append(configuration.file_path)
                self.profile_file_types[configuration.file_path] = AlphaCore
                if configuration.data.os is not None:
                    self.defined_os.add(configuration.data.os.name)
            elif isinstance(configuration.data, AlphaManifest):
                for profile in configuration.data.profiles:
                    dependencies = self.profile_dependencies[profile.name]
                    for dependency in profile.dependencies:
                        dependencies.append(dependency)
                self.profile_file_types[configuration.file_path] = AlphaManifest
            elif isinstance(configuration.data, AlphaDeveloperEnvironments):
                self.dev_environments[configuration.file_path] = configuration.data
                self.profile_map[configuration.data.profile].append(configuration.file_path)
                self.profile_file_types[configuration.file_path] = AlphaDeveloperEnvironments
            elif isinstance(configuration.data, AlphaDefaultSettings):
                logger.error("FIXME THIS SHOULDN'T EXIST")
                self.default_settings[configuration.file_path] = configuration.data
                self.profile_file_types[configuration.file_path] = AlphaDefaultSettings

    def build(self, profiles: Optional[list[str]], specified_oses: Optional[list[str]]) -> Dict[str, Path]:
        # Clears out the current build directory if it already exists
        if self.output_path.exists():
            shutil_rmtree(self.output_path)
        self.output_path.mkdir(parents=True, exist_ok=True)

        active_profiles = self.__get_active_profiles(profiles)
        package_paths: Dict[str, Path] = {}

        for active_os in self.__get_active_os(specified_oses):
            for active_profile in active_profiles:
                build_data = self.__get_build_files(active_profile, active_os)
                package_paths[f"{active_os}.{active_profile}"] = self.__build_package(build_data, active_os, active_profile)
        return package_paths

    def __get_active_profiles(self, profiles: Optional[list[str]]) -> set[str]:
        # When no profiles are passed in we generate all of them
        if profiles is None:
            # .key() only returns a set-like object, not a full set object with a .intersection method,
            # so we wrap the iterable returned in a set
            filtered_profiles = set(self.profile_dependencies.keys())
        else:
            filtered_profiles = set(profiles)
        active_profiles = filtered_profiles.intersection(self.profile_dependencies.keys())
        logger.debug(f"Generating build packages for the following active profiles: {active_profiles}")
        return active_profiles

    def __get_active_os(self, specified_oses: Optional[list[str]]) -> set[OSName]:
        # When no OSes are passed in we generate all of them
        if specified_oses is None:
            filtered_oses = self.defined_os
        else:
            filtered_oses = set([OSName[specified_os] for specified_os in specified_oses])
        active_oses = filtered_oses.intersection(self.defined_os)
        logger.debug(f"Building the following OSes: {active_oses}")
        return active_oses

    ##
    # WARNING: Everything below this line is terrible 6am code, come back and refactor after
    ##

    def __get_build_files(self, profile: str, active_os: OSName) -> list[Configuration]:
        dependency_paths = self.__map_profile_to_dependencies(profile)
        dependencies = []
        for dependency_path in dependency_paths:
            dependency_type = self.profile_file_types.get(dependency_path)
            if dependency_type == AlphaCore:
                core_data = self.core.get(dependency_path)
                if core_data.os is None or core_data.os.name is active_os:
                    logger.debug(
                        f"Profile {profile} on {active_os.value} has the following dependency path: {dependency_path}")
                    dependencies.append(Configuration(dependency_path, core_data))
            elif dependency_type == AlphaDeveloperEnvironments:
                developer_environment_data = self.dev_environments.get(dependency_path)
                logger.debug(
                    f"Profile {profile} on {active_os.value} has the following dependency path: {dependency_path}")
                dependencies.append(Configuration(dependency_path, developer_environment_data))
        return dependencies

    def __map_profile_to_dependencies(self, profile: str) -> list[Path]:
        combined_dependencies = []
        for dependency in self.profile_dependencies[profile]:
            for dependency_path in self.profile_map[dependency]:
                combined_dependencies.append(dependency_path)
        return combined_dependencies

    def __build_package(self, build_data: list[Configuration], active_os: OSName, active_profile: str) -> Path:
        dirpath = self.output_path.joinpath(f"dotfiles-{active_os.value}-{active_profile}")
        dirpath.mkdir(parents=True, exist_ok=True)

        # The following line when uncommented will write to a temp directory instead
        # dirpath = Path(mktempdir(None, f"pydotfiles-{datetime.now().isoformat()}-{active_os.value}-{active_profile}-"))

        is_installation_init_written = False
        for build_datum in build_data:
            package_builder = PackageBuilder(build_datum.data, build_datum.file_path, dirpath, self.base_dir, active_os)

            if not is_installation_init_written:
                package_builder.write_init()
                is_installation_init_written = True
            package_builder.build()
            package_builder.write_tail()

            if self.is_release:
                self.__zip_dir(dirpath)
        return dirpath

    def __consolidate_requirements(self, build_data: list[Configuration]) -> list[Configuration]:
        """
        Note: not working, pathing gets messed up depending on load order, so avoid this for now

        One thing that we'd like to do is to consolidate any configurations such as OS-level/package-level
        configurations so that they can occur all at once, instead of sporadically.
        """
        final_build_data = []
        os_build_data = []

        for build_datum in build_data:
            if isinstance(build_datum.data, AlphaCore):
                if build_datum.data.os is None:
                    final_build_data.append(build_datum)
                else:
                    os_build_data.append(build_datum)

        maybe_consolidated_data = self.__consolidate_os_build_data(os_build_data)
        if maybe_consolidated_data is not None:
            final_build_data.append(maybe_consolidated_data)
        return final_build_data

    def __consolidate_os_build_data(self, build_data: list[Configuration]) -> Optional[Configuration]:
        if len(build_data) == 0:
            return None

        if len(build_data) == 1:
            return build_data[0]

        consolidated_data = {"profile": "consolidated("}
        consolidated_actions = []
        consolidated_os = {
            "packages": [],
            "applications": [],
            "default_dock": [],
            "default_settings_files": [],
        }

        for build_datum in build_data:
            data: AlphaCore = build_datum.data
            consolidated_data["version"] = data.version
            consolidated_data["schema"] = data.schema

            # In reality this value isn't used in later code, so it's fine if this is inaccurate
            consolidated_data["profile"] += data.profile

            for action in data.actions:
                consolidated_actions.append(action.to_dict())

            original_os: OperatingSystem = data.os
            consolidated_os["name"] = original_os.name.value

            if original_os.shell is not None and consolidated_os.get("shell") is None:
                consolidated_os["shell"] = original_os.shell.value

            for package in original_os.packages:
                consolidated_os["packages"].append(package)

            for application in original_os.applications:
                consolidated_os["applications"].append(application)

            for default_dock_item in original_os.default_dock:
                consolidated_os["default_dock"].append(default_dock_item)

            for default_settings_file in original_os.default_settings_files:
                consolidated_os["default_settings_files"].append(default_settings_file)

        consolidated_data["profile"] += ")"
        consolidated_data["actions"] = consolidated_actions
        consolidated_data["os"] = consolidated_os

        return Configuration(
            file_path=build_data[0].file_path,
            data=AlphaCore.from_dict(consolidated_data)
        )

    def __zip_dir(self, directory: Path) -> Path:
        destination = Path(f"{directory}.zip")
        with ZipFile(destination, mode='w') as zip_fp:
            for content in directory.iterdir():
                zip_fp.write(content, content.name)
        return destination

##
# Build package methods
##


@dataclass(frozen=True, order=True)
class Symlink:
    initial_origin_file: Path
    destination_in_build_package: Path
    script_origin_file: Path
    script_destination_file: Path
    is_sudo: bool


@dataclass(frozen=True, order=True)
class Copy:
    initial_origin_file: Path
    destination_in_build_package: Path
    script_origin_file: Path
    script_destination_file: Path
    is_sudo: bool


@dataclass(frozen=True, order=True)
class RunScript:
    initial_origin_file: Path
    destination_in_build_package: Path
    script_origin_file: Path
    # No need for script destination file since the command is just being executed,
    # and not copied or symlinked elsewhere
    is_sudo: bool


class PackageBuilder:
    def __init__(self, data: AlphaCore | AlphaDeveloperEnvironments, file_path: Path, dirpath: Path, base_dir: Path, active_os: OSName):
        self.data = data
        self.file_path = file_path
        self.dirpath = dirpath
        self.installation_script = Path.joinpath(dirpath, "install.sh")
        self.base_dir = base_dir
        self.active_os = active_os

    def write_init(self) -> None:
        with self.installation_script.open('a') as init_fp:
            init_fp.write(f"#!/usr/bin/env bash\n\n# Fails on the first error\nset -e\n\n{template_starter_helper_functions()}")
        os_chmod(self.installation_script, 0o744)

    def write_tail(self) -> None:
        with self.installation_script.open('a') as init_fp:
            init_fp.write("\n")

    def build(self):
        if isinstance(self.data, AlphaCore):
            self.__add_to_build_package_core(self.data)
        elif isinstance(self.data, AlphaDeveloperEnvironments):
            self.__add_to_build_package_developer_env(self.data)

    def __add_to_build_package_core(self, data: AlphaCore) -> None:
        if data.os is not None:
            self.__add_os_to_build_package(data.os)
        self.__add_actions_to_build_package(data.actions)

    def __add_os_to_build_package(self, data: OperatingSystem) -> None:
        self.__add_os_to_install_script(data)

    def __add_os_to_install_script(self, data: OperatingSystem) -> None:
        os_instructions = ""
        os_instructions += template_os_packages(data)
        os_instructions += template_os_applications(data)
        os_instructions += template_os_default_dock(data)
        os_instructions += template_os_default_setting_files(data)

        with self.installation_script.open('a') as install_fp:
            install_fp.write(os_instructions)

    def __add_actions_to_build_package(self, data: list[ActionType]) -> None:
        for datum in data:
            self.__add_action_to_build_package(datum)

    def __add_action_to_build_package(self, data: ActionType) -> None:
        action_switcher = {
            Action.copy: self.__add_copy_files,
            Action.script: self.__add_run_scripts,
            Action.symlink: self.__add_symlinks,
        }

        action_switcher.get(data.action)(data)

    def __add_copy_files(self, data: ActionType) -> None:
        copy_actions = self.__get_copy_actions(data)
        # for copy_action in copy_actions:
        #     logger.debug(f"Adding in copy\n--------\n{copy_action}")
        self.__copy_files_to_build_package(copy_actions)
        self.__add_copy_instruction_to_install_script(copy_actions)

    def __get_copy_actions(self, data: ActionType) -> Set[Copy]:
        files_to_copy = set()
        for file_name, destination_name in data.files.items():
            if file_name == "*":
                files_to_copy = self.__get_files_to_copy(self.file_path.parent, Path(destination_name), data.sudo, data.hidden) | files_to_copy
            else:
                if data.absolute:
                    files_to_copy.add(Copy(
                        initial_origin_file=None,
                        destination_in_build_package=None,
                        script_origin_file=Path(file_name),
                        script_destination_file=Path(destination_name),
                        is_sudo=data.sudo
                    ))
                    continue

                files_to_copy.add(Copy(
                    initial_origin_file=self.file_path.parent.joinpath(Path(file_name)),
                    destination_in_build_package=self.dirpath.joinpath(
                        self.file_path.parent.joinpath(Path(file_name)).relative_to(self.base_dir)),
                    script_origin_file=self.file_path.parent.joinpath(Path(file_name)).relative_to(self.base_dir),
                    script_destination_file=Path(destination_name),
                    is_sudo=data.sudo
                ))
        return files_to_copy

    def __get_files_to_copy(self, directory: Path, destination_dir: Path, is_sudo: bool, is_hidden: bool) -> Set[Copy]:
        files_to_copy = set()
        for path_prefix, directory_names, file_names in os_walk(directory):
            for file_name_path in file_names:
                if file_name_path.endswith(".json") or file_name_path.endswith(".yaml") or file_name_path.endswith(".yml") or file_name_path.endswith(".symlink"):
                    continue

                destination_file = Path(f"{'.' if is_hidden else ''}{file_name_path}")
                files_to_copy.add(Copy(
                    initial_origin_file=Path(os_path.join(path_prefix, file_name_path)),
                    destination_in_build_package=self.dirpath.joinpath(
                        Path(os_path.join(path_prefix, file_name_path)).relative_to(self.base_dir)),
                    script_origin_file=Path(os_path.join(path_prefix, file_name_path)).relative_to(self.base_dir),
                    script_destination_file=destination_dir.joinpath(destination_file),
                    is_sudo=is_sudo
                ))
        return files_to_copy

    def __add_copy_instruction_to_install_script(self, copies: set[Copy]) -> None:
        copy_instruction = ""
        for copy in copies:
            copy_instruction += template_copy_file(copy.script_origin_file, copy.script_destination_file)
        with self.installation_script.open('a') as install_fp:
            install_fp.write(copy_instruction)

    def __add_run_scripts(self, data: ActionType) -> None:
        run_scripts = self.__get_runscripts(data)
        for run_script in run_scripts:
            logger.debug(f"Adding in run script:\n---------------{run_script}---------------\n")
        self.__copy_files_to_build_package(run_scripts)
        self.__add_runscript_instructions_to_install_script(run_scripts)

    def __get_runscripts(self, data: ActionType) -> set[RunScript]:
        files_to_execute = set()
        for file_name, destination_name in data.files.items():
            if file_name == "*":
                # In group symlinks, destination_name is a directory
                wildcard_scripts = self.__get_files_to_execute(self.file_path.parent, Path(destination_name), data.sudo, data.hidden)
                files_to_execute = wildcard_scripts | files_to_execute
            else:
                if data.absolute:
                    files_to_execute.add(RunScript(
                        initial_origin_file=None,
                        destination_in_build_package=None,
                        script_origin_file=Path(file_name),
                        is_sudo=data.sudo
                    ))
                    continue

                files_to_execute.add(RunScript(
                    initial_origin_file=self.file_path.parent.joinpath(Path(file_name)),
                    destination_in_build_package=self.dirpath.joinpath(
                        self.file_path.parent.joinpath(Path(file_name)).relative_to(self.base_dir)),
                    script_origin_file=self.file_path.parent.joinpath(Path(file_name)).relative_to(self.base_dir),
                    is_sudo=data.sudo
                ))

        # logger.debug(f"The script files to execute are: {files_to_execute}")
        return files_to_execute

    def __get_files_to_execute(self, directory: Path, destination_dir: Path, is_sudo: bool, is_hidden: bool) -> Set[RunScript]:
        files_to_execute = set()
        for path_prefix, directory_names, file_names in os_walk(directory):
            for file_name_path in file_names:
                # We cheat here a bit and say "if you're specifying a wildcard execute script operation,
                # then every file in this directory that's not a symlink or copy operation can be executed,
                # and we don't check for whether the executable bit (since we're lazy, we can add it in later
                # if it turns out useful
                if not file_name_path.endswith(".symlink") and not file_name_path.endswith(".json") and not file_name_path.endswith(".yaml") and not file_name_path.endswith(".yml"):
                    files_to_execute.add(RunScript(
                        initial_origin_file=Path(os_path.join(path_prefix, file_name_path)),
                        destination_in_build_package=self.dirpath.joinpath(Path(os_path.join(path_prefix, file_name_path)).relative_to(self.base_dir)),
                        script_origin_file=Path(os_path.join(path_prefix, file_name_path)).relative_to(self.base_dir),
                        is_sudo=is_sudo
                    ))
        return files_to_execute

    def __add_runscript_instructions_to_install_script(self, run_scripts: set[RunScript]) -> None:
        script_instruction = ""
        for run_script in run_scripts:
            script_instruction += template_run_script(run_script.script_origin_file, run_script.is_sudo)
        with self.installation_script.open('a') as install_fp:
            install_fp.write(script_instruction)

    def __add_symlinks(self, data: ActionType) -> None:
        symlinks = self.__get_symlinks(data)
        # for symlink in symlinks:
        #     logger.debug(f"Adding in symlink\n--------\n{symlink}")
        self.__copy_files_to_build_package(symlinks)
        self.__add_symlink_instruction_to_install_script(symlinks)

    def __get_symlinks(self, data: ActionType) -> set[Symlink]:
        files_to_symlink = set()
        for file_name, destination_name in data.files.items():
            if file_name == "*":
                # In group symlinks, destination_name is a directory
                wildcard_symlinks = self.__get_files_to_symlink(self.file_path.parent, Path(destination_name), data.sudo, data.hidden)
                files_to_symlink = wildcard_symlinks | files_to_symlink
            else:
                if data.absolute:
                    files_to_symlink.add(Symlink(
                        initial_origin_file=None,
                        destination_in_build_package=None,
                        script_origin_file=Path(file_name),
                        script_destination_file=Path(destination_name),
                        is_sudo=data.sudo
                    ))
                    continue

                files_to_symlink.add(Symlink(
                    initial_origin_file=self.file_path.parent.joinpath(Path(file_name)),
                    destination_in_build_package=self.dirpath.joinpath(
                        self.file_path.parent.joinpath(Path(file_name)).relative_to(self.base_dir)),
                    script_origin_file=self.file_path.parent.joinpath(Path(file_name)).relative_to(self.base_dir),
                    script_destination_file=Path(destination_name),
                    is_sudo=data.sudo
                ))

        # logger.debug(f"The files to symlink are: {files_to_symlink}")
        return files_to_symlink

    def __get_files_to_symlink(self, directory: Path, destination_dir: Path, is_sudo: bool, is_hidden: bool) -> Set[Symlink]:
        files_to_symlink = set()
        for path_prefix, directory_names, file_names in os_walk(directory):
            for file_name_path in file_names:
                if file_name_path.endswith(".symlink"):
                    destination_file = Path(f"{'.' if is_hidden else ''}{file_name_path}".replace(".symlink", ""))
                    files_to_symlink.add(Symlink(
                        initial_origin_file=Path(os_path.join(path_prefix, file_name_path)),
                        destination_in_build_package=self.dirpath.joinpath(Path(os_path.join(path_prefix, file_name_path)).relative_to(self.base_dir)),
                        script_origin_file=Path(os_path.join(path_prefix, file_name_path)).relative_to(self.base_dir),
                        script_destination_file=destination_dir.joinpath(destination_file),
                        is_sudo=is_sudo
                    ))
        return files_to_symlink

    def __copy_files_to_build_package(self, actions: set[Copy] | set[Symlink] | set[RunScript]) -> None:
        for action in actions:
            if action.initial_origin_file is None:
                logger.debug(f"Skipping file copying since no file was given (absolute path)")
                continue
            logger.debug(f"Copying the file {action.initial_origin_file} to {action.destination_in_build_package}")
            try:
                action.destination_in_build_package.parent.mkdir(parents=True, exist_ok=True)
                shutil_copy2(action.initial_origin_file, action.destination_in_build_package)
            except FileNotFoundError:
                logger.exception(f"The action was: {action}")

    def __add_symlink_instruction_to_install_script(self, symlinks: set[Symlink]) -> None:
        symlink_instruction = ""
        for symlink in symlinks:
            symlink_instruction += template_symlink_file(symlink.script_origin_file, symlink.script_destination_file, symlink.is_sudo)
        with self.installation_script.open('a') as install_fp:
            install_fp.write(symlink_instruction)

    def __add_to_build_package_developer_env(self, data: AlphaDeveloperEnvironments) -> None:
        for environment in data.environments:
            self.__add_dev_environment_to_install_script(environment)

    def __add_dev_environment_to_install_script(self, data: Environment) -> None:
        dev_environment_setup = template_dev_environment_setup(data, self.active_os)
        with self.installation_script.open('a') as install_fp:
            install_fp.write(dev_environment_setup)



# General imports
from argparse import ArgumentParser, RawDescriptionHelpFormatter
from pathlib import Path
from os import getcwd
from logging import getLogger
from json import dumps as json_dumps

# Project imports
from .deprecated import ArgumentDispatcher
from pydotfiles.v4.builder import Builder
from pydotfiles.v4.loader import load_configuration
from pydotfiles import __version__ as version
from pydotfiles.v4.logging import set_logging
from pydotfiles.v4.common import PydotfilesError
from pydotfiles.v4.logging import PrettyPrint

logger = getLogger(__name__)


class ArgumentDispatcherFacade:
    def __init__(self, api_arguments: list[str]):
        self.api_arguments = api_arguments
        self.deprecated_dispatcher = ArgumentDispatcher(api_arguments)

    def dispatch(self) -> None:
        valid_commands = {
            'download',
            'install',
            'uninstall',
            'update',
            'clean',
            'set',
            'validate',
            'build',
        }

        parser = ArgumentParser(formatter_class=RawDescriptionHelpFormatter, description="""
        Python Dotfiles Manager, enabling configuration-based management of your system!

        Commands:
          - download: Downloads your dotfiles onto your computer
          - install: Installs all/part of your dotfiles
          - uninstall: Uninstalls all/part of your dotfiles
          - update: Updates all/part of your dotfiles
          - clean: Removes the pydotfiles cache/default
          - set: Sets configuration values for managing your dotfiles
          - validate: Validates that a given directory is pydotfiles-compliant
          - build: Generates dotfile profile builder packages (can be used as part of a CI/CD pipeline)
        """)
        parser.add_argument('--version', action='version', version='%(prog)s ' + version)
        parser.add_argument("command", help="Runs the command given", choices=valid_commands)

        command = self.api_arguments[1:2]
        command_arguments = self.api_arguments[2:]

        args = parser.parse_args(command)

        # Dynamically dispatches to the relevant method
        if command[0] == "build":
            getattr(self, args.command)(command_arguments)
            return

        try:
            getattr(self.deprecated_dispatcher, args.command)(command_arguments)
        except PydotfilesError as e:
            logger.exception(f"an error occurred: {e.context_map}")

    def build(self, command_arguments: list[str]) -> None:
        help_description = """
        Builds any/all profiles into self-contained builder packages, which can be easily downloaded
        and installed without having to install/run pydotfiles.

        (default: Checks the current working directory)
        """
        parser = self.__get_base_parser(help_description, "build")
        parser.add_argument("-d", "--directory", help="Builds all/indicated profile build packages located in the passed-in directory. Defaults to the current working directory", default=getcwd())
        parser.add_argument("-b", "--build-directory", help="The output directory. Defaults to pydotfiles-build in the current working directory", default=str(Path(getcwd()).joinpath("pydotfiles-build")))
        parser.add_argument("-r", "--release", help="Will generate releases (.zip) of each of the build files", action="store_true")

        parser.add_argument("-p", "--profiles", nargs='+', help="Indicates which profiles should be built, and how they should be composed. Defaults to all profiles being built independently")
        parser.add_argument("-o", "--operating-systems", nargs='+', help="Indicates which operating system build packages should be built. Defaults to all profiles being built independently")

        args = parser.parse_args(command_arguments)

        set_logging(args.quiet, args.verbose)

        configurations = load_configuration(Path(args.directory))

        builder = Builder(configurations, Path(args.directory), Path(args.build_directory), args.release)
        build_packages = builder.build(args.profiles, args.operating_systems)
        for name, build_package_path in build_packages.items():
            PrettyPrint.success(f"[{name}]\tSuccessfully generated build package in: {build_package_path}")

    @staticmethod
    def __get_base_parser(description: str, sub_command: str) -> ArgumentParser:
        parser = ArgumentParser(
            prog=f"pydotfiles {sub_command}",
            formatter_class=RawDescriptionHelpFormatter,
            description=description
        )
        logging_parser_group = parser.add_mutually_exclusive_group()
        logging_parser_group.add_argument("-v", "--verbose", help="Enables more verbose logging", action="store_true")
        logging_parser_group.add_argument("-q", "--quiet", help="Squelches the default logging (still outputs to stderr upon failures)", action="store_true")
        return parser



# General imports
from typing import Dict, Optional
from dataclasses import dataclass
from enum import Enum

# Project imports
from .os import OperatingSystem, OSName
from .action import ActionType
from .environment import Environment
from .default_settings import DefaultSettings
from .profiles import Profile


class Version(Enum):
    alpha = "alpha"


class Schema(Enum):
    core = "core"
    default_settings = "default_settings"
    developer_environments = "developer_environments"
    manifest = "manifest"


@dataclass(frozen=True, order=True)
class AlphaCommon:
    version: Version
    schema: Schema


@dataclass(frozen=True, order=True)
class AlphaCore(AlphaCommon):
    profile: str
    os: Optional[OperatingSystem]
    actions: list[ActionType]

    @staticmethod
    def from_dict(data: Dict) -> 'AlphaCore':
        return AlphaCore(
            version=data.get("version"),
            schema=data.get("schema"),
            profile=data.get("profile"),
            os=OperatingSystem.from_dict(data.get("os")),
            actions=ActionType.from_dict(data.get("actions", []))
        )

    def join_with_default_settings_data(self, additional_data: list[DefaultSettings]) -> 'AlphaCore':
        return AlphaCore(
            version=self.version,
            schema=self.schema,
            profile=self.profile,
            os=self.os.join_with_default_settings_data(additional_data),
            actions=self.actions,
        )


@dataclass(frozen=True, order=True)
class AlphaDefaultSettings(AlphaCommon):
    default_settings: list[DefaultSettings]

    @staticmethod
    def from_dict(data: Dict) -> 'AlphaDefaultSettings':
        return AlphaDefaultSettings(
            version=data.get("version"),
            schema=data.get("schema"),
            default_settings=DefaultSettings.from_dict(data.get("default_settings", []))
        )


@dataclass(frozen=True, order=True)
class AlphaDeveloperEnvironments(AlphaCommon):
    environments: list[Environment]
    profile: str

    @staticmethod
    def from_dict(data: Dict) -> 'AlphaDeveloperEnvironments':
        return AlphaDeveloperEnvironments(
            version=data.get("version"),
            schema=data.get("schema"),
            profile=data.get("profile"),
            environments=Environment.from_dict(data.get("environments", []))
        )


@dataclass(frozen=True, order=True)
class AlphaManifest(AlphaCommon):
    profiles: list[Profile]

    @staticmethod
    def from_dict(data: Dict) -> 'AlphaManifest':
        return AlphaManifest(
            version=data.get("version"),
            schema=data.get("schema"),
            profiles=Profile.from_dict(data.get("profiles"))
        )


def get_schema(schema: str) -> type:
    configuration_map: Dict[str, type] = {
        "core": AlphaCore,
        "default_settings": AlphaDefaultSettings,
        "developer_environments": AlphaDeveloperEnvironments,
        "manifest": AlphaManifest,
    }
    return configuration_map.get(schema)



# General imports
import logging
import shutil
import subprocess
import getpass
import sys
import json
import yaml
from pathlib import Path

# Project imports
from pydotfiles.models import OverrideAction
from pydotfiles.utils import PrettyLogFormatter, PrettyPrint


logger = logging.getLogger(__name__)


"""
Project-specific logic
"""


def install_homebrew():
    if shutil.which("brew") is None:
        logger.info(f"Package Manager: Homebrew was not found, installing now (this may take a while)")
        command = "/usr/bin/ruby -e \"$(curl -fsSL https://raw.githubusercontent.com/Homebrew/install/master/install)\""

        process = subprocess.Popen(command.split(), stdin=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
        try:
            process.communicate('\r')
            logger.info(f"Package Manager: Successfully installed homebrew")
        except subprocess.TimeoutExpired:
            process.kill()
            logger.exception(f"Package Manager: Failed to install homebrew")
            raise
    else:
        logger.info(f"Package Manager: Homebrew is already installed")


def uninstall_homebrew():
    if shutil.which("brew") is None:
        logger.info(f"Package Manager: Homebrew is already uninstalled")
    else:
        logger.info(f"Package Manager: Uninstalling homebrew (this may take a while)")
        command = "ruby -e \"$(curl -fsSL https://raw.githubusercontent.com/Homebrew/install/master/uninstall)\""

        process = subprocess.Popen(command.split(), stdin=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
        logger.info(f"Package Manager: Successfully uninstalled homebrew")

        # TODO P2: implement
        # It's unknown what parameters we need to pass into the script, so comment this out for now
        # try:
        #     process.communicate('\r')
        # except TimeoutExpired:
        #     process.kill()
        #     PrettyPrint.fail(f"Package Manager: Failed to install homebrew")
        #     raise


def load_data_from_file(config_file):
    """
    Loads a configuration file
    """
    if config_file is None:
        return {}

    # TODO P3: We're doing this path -> string convert here to not break
    # current APIs, but we'll want to come back and re-write everything in
    # terms of Path later on
    if isinstance(config_file, Path):
        config_file = str(config_file)

    with open(config_file, 'r') as config_fd:
        if config_file.endswith("json"):
            return json.load(config_fd)
        elif config_file.endswith("yaml") or config_file.endswith("yml"):
            try:
                return yaml.load(config_fd)
            except yaml.YAMLError:
                raise
        else:
            raise RuntimeError(f"Configuration Data Load: The file type of the settings configuration file {config_file} could not be parsed (not a supported filetype)")


"""
User-interfacing functions
"""


def ask_sudo_password():
    PrettyPrint.user(f"Sudo: Actions requiring sudo actions have been detected, please enter your password below (your password is NOT stored at any point in time)")
    sudo_password = getpass.getpass()
    return sudo_password


def get_user_override(action):
    """
    When an issue occurs where the destination of
    a FileAction already exists, we ask the user
    what to do
    """
    while True:
        decision = input(f"Action: File already exists, what would you like to do? [Destination={action.destination}]\n[s]kip, [S]kip all, [o]verwrite, [O]verwrite all, [b]ackup, [B]ackup all\n> ")

        if decision == 's':
            return OverrideAction.SKIP_FILE
        elif decision == 'S':
            return OverrideAction.SKIP_ALL_FILES
        elif decision == 'o':
            return OverrideAction.OVERWRITE_FILE
        elif decision == 'O':
            return OverrideAction.OVERWRITE_ALL_FILES
        elif decision == 'b':
            return OverrideAction.BACKUP_FILE
        elif decision == 'B':
            return OverrideAction.BACKUP_ALL_FILES


"""
General utils
"""


def set_logging(is_quiet, is_verbose):
    if is_quiet:
        logging_level = logging.WARNING
    elif is_verbose:
        logging_level = logging.DEBUG
    else:
        logging_level = logging.INFO

    # Enables normal logging to stdout and errors to stderr
    stderr_handler = logging.StreamHandler(sys.stderr)
    stderr_handler.setLevel(logging.WARNING)
    stderr_handler.setFormatter(PrettyLogFormatter())
    logging.root.addHandler(stderr_handler)

    # See https://github.com/Robpol86/sphinxcontrib-versioning/blob/master/sphinxcontrib/versioning/setup_logging.py
    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setLevel(logging_level)
    stdout_handler.setFormatter(PrettyLogFormatter())
    stdout_handler.addFilter(type('', (logging.Filter,), {'filter': staticmethod(lambda r: r.levelno < logging.WARNING)}))
    logging.root.addHandler(stdout_handler)

    logging.root.setLevel(logging_level)



# General imports
import configparser
import shutil
import os
import subprocess
import sys
import re
from subprocess import Popen, PIPE, TimeoutExpired
from distutils.version import StrictVersion
import logging

# Vendor imports
from git import Repo
from git.remote import RemoteProgress
from progressbar import ProgressBar

# Project imports
from .enums import FileActionType, OverrideAction
from .constants import *
from .primitives import FileAction, CacheDirectory
from .exceptions import PydotfilesError, PydotfilesErrorReason, ValidationError
from .validator import Validator

from .utils import install_homebrew, uninstall_homebrew, load_data_from_file
from .utils import get_user_override, ask_sudo_password
from .dock import DockManager
from pydotfiles.utils import remove_prefix
from pydotfiles.defaults import get_current_mac_version
from .utils import set_logging
from pydotfiles.loading import get_os_default_settings
from pydotfiles.common import OS, PackageManager
from pydotfiles.loading import parse_developer_environments


logger = logging.getLogger(__name__)


class Dotfiles:
    """
    Class representing the dotfiles that
    we are reading in
    """

    def __init__(self, config_repo_local, config_repo_remote, is_quiet, is_verbose, active_modules=None):
        self.config_repo_local = config_repo_local
        self.host_os = OS.from_string(sys.platform)
        self.cache_directory = CacheDirectory(package_manager=OS.get_package_manager(self.host_os))

        set_logging(is_quiet, is_verbose)

        if config_repo_remote is None:
            self.config_repo_remote = load_config_repo_remote(config_repo_local) if self.is_cloned else None
        else:
            self.config_repo_remote = config_repo_remote

        module_information = load_active_modules(config_repo_local, active_modules, self.host_os, self.cache_directory) if self.is_cloned else None

        self.modules = None if module_information is None else module_information[0]
        self.is_sudo_used = None if module_information is None else module_information[1]
        self.sudo_password = None

    def __str__(self):
        return f"Dotfiles [Config Directory={self.config_repo_local}, Config Remote={self.config_repo_remote}, Modules={self.modules}]"

    @property
    def is_cloned(self):
        return os.path.isdir(f"{self.config_repo_local}/.git")

    def download(self):
        """
        Dotfile configs haven't been cloned yet, so cloning
        from remote to local
        """
        if self.config_repo_remote is None:
            raise PydotfilesError(PydotfilesErrorReason.NO_REMOTE_REPO, "Download: No dotfiles configuration git remote link passed in (try running `pydotfiles download -r <git remote link>`)")

        logger.info(f"Download: Cloning git repository [remote={self.config_repo_remote}, local={self.config_repo_local}]")
        try:
            Repo.clone_from(self.config_repo_remote, self.config_repo_local, progress=GitRemoteProgress())
            logger.info(f"Download: Successfully cloned git repository [remote={self.config_repo_remote}, local={self.config_repo_local}]")
        except Exception:
            logger.exception(f"Download: Failed to clone git repository [Remote={self.config_repo_remote}, Local={self.config_repo_local}]")
            raise PydotfilesError(PydotfilesErrorReason.REMOTE_REPO_CLONE_ISSUE)

    def clean(self, target):
        """
        Deletes the downloaded dotfile configs directory
        """
        if target == "cache":
            if os.path.isdir(PYDOTFILES_CACHE_DIRECTORY):
                logger.info(f"Clean: Deleting the pydotfiles cache [Cache={PYDOTFILES_CACHE_DIRECTORY}]")
                shutil.rmtree(PYDOTFILES_CACHE_DIRECTORY)
                logger.info(f"Clean: Successfully deleted the pydotfiles cache [Cache={PYDOTFILES_CACHE_DIRECTORY}]")
            else:
                logger.info(f"Clean: The pydotfiles cache does not exist [Cache={PYDOTFILES_CACHE_DIRECTORY}]")
        elif target == "repo":
            if os.path.isdir(self.config_repo_local):
                logger.info(f"Clean: Deleting the pydotfiles configuration local repo [Local={self.config_repo_local}]")
                shutil.rmtree(self.config_repo_local)
                logger.info(f"Clean: Successfully deleted the pydotfiles configuration local repo [Local={self.config_repo_local}]")
            else:
                logger.info(f"Clean: The pydotfiles configuration local repo does not exist [Local={self.config_repo_local}]")
        else:
            raise PydotfilesError(PydotfilesErrorReason.UNKNOWN_CLEANING_TARGET, f"Clean: Unknown cleaning target passed in [target={target}]")

    """
    Installation methods
    """

    def install_all(self):
        # Propagates the sudo password for the program
        self.sudo_password = ask_sudo_password() if self.is_sudo_used else None
        self.__propagate_sudo_password__()

        for module_name, module in self.modules.items():
            module.install()

    def install_multiple_modules(self, module_names):
        # Propagates the sudo password for the program
        self.sudo_password = ask_sudo_password() if self.is_sudo_used else None
        self.__propagate_sudo_password__()

        for module_name in module_names:
            self.install_single_module(module_name)

    def install_single_module(self, module_name):
        module = self.modules.get(module_name)

        if module is None:
            raise PydotfilesError(PydotfilesErrorReason.UNKNOWN_MODULE_NAME, f"Module Installation: Failed to find the module `{module_name}` when trying to install")

        module.install()

    """
    Uninstallation methods
    """

    def uninstall_all(self, uninstall_packages, uninstall_applications, uninstall_environments):
        # Propagates the sudo password for the program
        self.sudo_password = ask_sudo_password() if self.is_sudo_used else None
        self.__propagate_sudo_password__()

        for module_name, module in self.modules.items():
            module.uninstall(uninstall_packages, uninstall_applications, uninstall_environments)

    def uninstall_multiple_modules(self, module_names, uninstall_packages, uninstall_applications, uninstall_environments):
        # Propagates the sudo password for the program
        self.sudo_password = ask_sudo_password() if self.is_sudo_used else None
        self.__propagate_sudo_password__()

        for module_name in module_names:
            self.uninstall_single_module(module_name, uninstall_packages, uninstall_applications, uninstall_environments)

    def uninstall_single_module(self, module_name, uninstall_packages, uninstall_applications, uninstall_environments):
        module = self.modules.get(module_name)
        if module is None:
            raise PydotfilesError(PydotfilesErrorReason.UNKNOWN_MODULE_NAME, f"Module Installation: Failed to find the module `{module_name}` when trying to uninstall")
        module.uninstall(uninstall_packages, uninstall_applications, uninstall_environments)

    def update(self):
        if self.config_repo_remote is None:
            raise PydotfilesError(PydotfilesErrorReason.NO_REMOTE_REPO, f"Update: No dotfiles configuration git remote link passed in")

        logger.info(f"Update: Pulling git repository [Remote={self.config_repo_remote}, Local={self.config_repo_local}]")
        try:
            Repo(self.config_repo_local).remote('origin').pull(progress=GitRemoteProgress())
            logger.info(f"Update: Successfully updated dotfiles git repository [Remote={self.config_repo_remote}, Local={self.config_repo_local}, Remote=Origin]")
        except Exception:
            logger.exception(f"Update: Failed to update git repository [Remote={self.config_repo_remote}, Local={self.config_repo_local}]")
            raise PydotfilesError(PydotfilesErrorReason.REMOTE_REPO_CLONE_ISSUE)

    def __propagate_sudo_password__(self):
        for module in self.modules.values():
            module.sudo_password = self.sudo_password
            module.__propagate_sudo_password__()


class Module:
    """
    Class representing a single module's configuration
    values
    """

    def __init__(self, name, directory, settings_file, start_file, post_file, undo_start_file, undo_post_file, symlinks, other_files, host_os, cache_directory):
        self.name = name
        self.directory = directory
        self.symlinks = symlinks
        self.other_files = other_files
        self.host_os = host_os
        self.cache_directory = cache_directory

        self.start_action = None if start_file is None else FileAction(FileActionType.SCRIPT, start_file, undo_start_file, None, None)
        self.post_action = None if post_file is None else FileAction(FileActionType.SCRIPT, post_file, undo_post_file, None, None)

        # Loads in the settings file
        self.settings_file = settings_file
        settings_data = load_data_from_file(self.settings_file)
        self.operating_system = parse_operating_system_config(settings_data.get('os'), self.cache_directory, self.directory)
        self.environments = parse_developer_environments(settings_data.get('environments'))
        self.actions, self.is_sudo_used = parse_action_configs(settings_data.get('actions'), self.directory, self.symlinks, self.other_files)
        self.sudo_password = None

    def __str__(self):
        return f"Module [Name={self.name}, Start={self.start_action}, Settings={self.settings_file}, Post={self.post_action}]"

    def install(self):
        if self.operating_system is not None and self.operating_system.name != self.host_os:
            logger.info(f"Install: Skipping operating system installation due to OS mismatch [HostOS={self.host_os}, ModuleOS={self.operating_system.name}]")
            return

        logger.info(f"Install: Installing module [name={self.name}]")

        if self.start_action is not None:
            self.start_action.do()

        if self.operating_system is not None:
            self.operating_system.install_packages()
            self.operating_system.install_applications()
            self.operating_system.install_settings()
            self.operating_system.install_default_dock()

        for environment in self.environments:
            environment.install()

        self.do_actions()

        if self.post_action is not None:
            self.post_action.do()

        logger.info(f"Install: Successfully installed module [name={self.name}]")

    def uninstall(self, uninstall_packages, uninstall_applications, uninstall_environments):
        if self.operating_system is not None and self.operating_system.name != self.host_os:
            logger.info(f"Uninstall: Skipping operating system uninstallation due to OS mismatch [HostOS={self.host_os}, ModuleOS={self.operating_system.name}]")
            return

        logger.info(f"Uninstall: Uninstalling module [name={self.name}]")

        if self.start_action is not None:
            self.start_action.undo()

        self.undo_actions()

        if uninstall_environments:
            for environment in self.environments:
                environment.uninstall()

        if self.operating_system is not None:
            if uninstall_applications:
                self.operating_system.uninstall_applications()

            if uninstall_packages:
                self.operating_system.uninstall_packages()

        if self.post_action is not None:
            self.post_action.undo()

        logger.info(f"Uninstall: Successfully uninstalled module [name={self.name}]")

    def do_actions(self):
        override_action = None
        for action in self.actions:
            try:
                override_action = self.__do_action_with_override__(override_action, action)
            except FileExistsError:
                logger.debug(f"Action: Failed to complete action [Action={action}]")
                override_action = get_user_override(action)
                override_action = self.__do_action_with_override__(override_action, action)

    def undo_actions(self):
        for file_action in self.actions:
            try:
                logger.info(f"Action: Starting undo action [Action={file_action}]")
                file_action.undo()
                logger.info(f"Action: Successfully undid action [Action={file_action}]")
            except Exception:
                logger.exception(f"Action: Failed to undo action [Action={file_action}]")
                raise

    @staticmethod
    def __do_action_with_override__(override_action, action):
        # Normal procedure
        if override_action is None:
            logger.info(f"Action: Starting action [Action={action}]")
            action.do()
            logger.info(f"Action: Successfully completed action [Action={action}]")
            return None

        # Skips the actions
        if override_action == OverrideAction.SKIP_FILE:
            logger.info(f"Action: Skipping action [Action={action}]")
            return None

        if override_action == OverrideAction.SKIP_ALL_FILES:
            logger.info(f"Action: Skipping all actions [Action={action}]")
            return OverrideAction.SKIP_ALL_FILES

        # Overwrites the actions
        if override_action == OverrideAction.OVERWRITE_FILE:
            logger.info(f"Action: Overwriting action destination [Action={action}]")
            action.overwrite()
            return None

        if override_action == OverrideAction.OVERWRITE_ALL_FILES:
            logger.info(f"Action: Overwriting all action's destinations [Action={action}]")
            action.overwrite()
            return OverrideAction.OVERWRITE_ALL_FILES

        # Backs up first before performing the action
        if override_action == OverrideAction.BACKUP_FILE:
            logger.info(f"Action: Backing up first for action [Action={action}]")
            action.backup()
            return None

        if override_action == OverrideAction.BACKUP_ALL_FILES:
            logger.info(f"Action: Backing up first for all actions [Action={action}]")
            action.backup()
            return None

    def __propagate_sudo_password__(self):
        for action in self.actions:
            action.sudo_password = self.sudo_password

        if self.operating_system is not None:
            self.operating_system.sudo_password = self.sudo_password

        if self.start_action is not None:
            self.start_action.sudo_password = self.sudo_password

        if self.post_action is not None:
            self.post_action.sudo_password = self.sudo_password


class OperatingSystem:
    """
    Class representing configurations for a given
    operating system
    """

    def __init__(self, name, shell, packages, applications, cache_directory, default_dock, settings, dock_manager):
        self.name = OS.from_string(name)
        self.shell = shell
        self.package_manager = OS.get_package_manager(self.name)
        self.packages = packages
        self.applications = applications
        self.cache_directory = cache_directory
        self.default_dock = default_dock
        self.settings = settings
        self.dock_manager = dock_manager
        self.sudo_password = None

    def install_package_manager(self):
        if self.name == OS.MACOS:
            install_homebrew()
        else:
            logger.warning(f"Package Manager: Support for package manager is not implemented yet [action=install, package manager={self.package_manager}]")

    def install_packages(self):
        for package in self.packages:
            self.install_package(package)

    def install_applications(self):
        for application in self.applications:
            self.install_application(application)

    def install_default_dock(self):
        if self.name != OS.MACOS:
            logger.info("Default Dock: Setting the dock is currently only supported on macOS- feel free to leave a github issue at https://github.com/JasonYao/pydotfiles/issues/new to request this feature!")
            return

        if self.default_dock is None:
            logger.info("Default Dock: No default dock specified, skipping dock setting")
            return

        current_dock_applications = self.dock_manager.get_current_dock_applications()

        # Note: The ^ gives the disjunctive union (symmetric difference) of two sets
        symmetric_difference_in_dock_applications = self.default_dock ^ current_dock_applications

        if len(symmetric_difference_in_dock_applications) == 0:
            logger.info("Default Dock: All default dock applications already set")
            return

        logger.info(f"Default Dock: Found persistent apps on the dock that did not belong or missing default apps, resetting now [default_apps_difference={symmetric_difference_in_dock_applications}]")

        # In this scenario, we found persistent apps on the dock that
        # don't belong there, or missing default apps, so we'll delete
        # the existing dock config, add in our new one, and then restart
        # everything
        DockManager.delete_dock_plist_file()

        for default_dock_application in self.default_dock:
            DockManager.add_application_to_default_on_dock(default_dock_application)

        DockManager.restart_dock()
        logger.info(f"Default Dock: Successfully set the dock to have the default applications [default_applications={self.default_dock}]")

    def install_settings(self):
        current_mac_version = get_current_mac_version()
        for setting in self.settings:
            if setting.should_run(current_mac_version, self.sudo_password):
                logger.info(f"Setting: Default setting not set, setting now [default_setting={setting.name}, description={setting.description}]")
                setting.run(self.sudo_password)
            else:
                logger.info(f"Setting: Default setting already set [default_setting={setting.name}, description={setting.description}]")

    def uninstall_package_manager(self):
        if self.name == OS.MACOS:
            uninstall_homebrew()
        else:
            logger.warning(f"Package Manager: Support for package manager is not implemented yet [action=uninstall, package manager={self.package_manager}]")

    def uninstall_packages(self):
        for application in self.applications:
            self.uninstall_application(application)

    def uninstall_applications(self):
        for application in self.applications:
            self.uninstall_application(application)

    """
    Helper methods
    """

    def install_package(self, package_with_args):
        package = package_with_args.split()[0]

        if self.cache_directory.is_package_installed(package):
            logger.info(f"Package: Already installed [Package={package}]")
            return

        logger.info(f"Package: Starting installation [Package={package}]")

        if self.package_manager == PackageManager.BREW:
            command = f"brew install {package_with_args}"

            command_result = subprocess.run(command.split(), stdout=subprocess.PIPE, stderr=subprocess.PIPE)

            # Version error-handling
            stderr = command_result.stderr.decode()
            if stderr == "":
                logger.info(f"Package: Successfully installed [Package={package}]")
            elif "is already installed and up-to-date" in stderr:
                logger.info(f"Package: Already installed and up to date [Package={package}]")
            elif "is already installed\n" in stderr:
                semver_extractor = re.compile("(\d+\.)(\d+\.?)(\d+\.?)?")
                all_semver = semver_extractor.findall(stderr)
                current_version = StrictVersion(''.join(all_semver[0]))
                latest_version = StrictVersion(''.join(all_semver[1]))
                logger.warning(f"Package: Installed but outdated [Package={package}, CurrentVersion={current_version}, LatestVersion={latest_version}, help=To update all your libraries, try running `pydotfiles update all` or `pydotfiles update {package}`]")
            else:
                raise NotImplementedError(f"Package: Failed to install package `{package}` [Stderr={stderr}]")

            # Dumps to the cache file after installation
            if os.path.isfile(self.cache_directory.package_cache_file):
                # Just appends the latest package to an existing cache
                self.cache_directory.append_package(package)
            else:
                logger.info(f"Caching: Dumping found packages into pydotfiles cache")
                # No cache was found, we do a full dump here at the start
                command = f"brew list"
                command_result = subprocess.run(command.split(), capture_output=True)
                package_dump = command_result.stdout.decode()
                self.cache_directory.overwrite_packages(package_dump)
                self.cache_directory.reload_packages()
        elif self.package_manager == PackageManager.APT or self.package_manager == PackageManager.YUM:
            # TODO P2: Clean up and fully flesh out support for *nix
            command = f"{self.package_manager} install -y {package}"

            process = Popen(['sudo', '-S'] + command.split(), stdin=PIPE, stderr=PIPE, universal_newlines=True)
            try:
                process.communicate(self.sudo_password + '\n')
            except TimeoutExpired:
                process.kill()
                logger.exception(f"Package: Installation failed [Package={package}]")
                raise

            logger.info(f"Package: Installation Successful [Package={package}]")

            # TODO P3: Put list of downloaded packages into cache
        else:
            raise NotImplementedError(f"Package: Install is currently not supported [PackageManager={self.package_manager}]")

    def install_application(self, application):
        if self.cache_directory.is_application_installed(application):
            logger.info(f"Application: Already installed [Application={application}]")
            return

        logger.info(f"Application: Starting installation [Application={application}]")

        if self.package_manager == PackageManager.BREW:
            command = f"brew cask install {application}"
            command_result = subprocess.run(command.split(), capture_output=True)

            stderr = command_result.stderr.decode()

            # Response code parsing
            if command_result.returncode == 0:
                logger.info(f"Application: Successfully installed [Application={application}]")
            elif "is already installed." in stderr:
                logger.info(f"Application: Already installed application [Application={application}]")
            else:
                logger.warning(f"Application: Failed to install [Application={application}, Stderr={stderr}]")

            # Dumps to the cache file after installation
            if os.path.isfile(self.cache_directory.application_cache_file):
                # Just appends the latest application to an existing cache
                self.cache_directory.append_application(application)
            else:
                logger.info(f"Caching: Dumping found applications into pydotfiles cache")
                # No cache was found, we do a full dump here at the start
                command = f"brew cask list"
                command_result = subprocess.run(command.split(), capture_output=True)
                applicatione_dump = command_result.stdout.decode()
                self.cache_directory.overwrite_applications(applicatione_dump)
                self.cache_directory.reload_applications()
        elif self.package_manager == PackageManager.APT or self.package_manager == PackageManager.YUM:
            # TODO P2: Clean up and fully flesh out support for *nix
            command = f"{self.package_manager} install -y {application}"

            process = Popen(['sudo', '-S'] + command.split(), stdin=PIPE, stderr=PIPE, universal_newlines=True)
            try:
                process.communicate(self.sudo_password + '\n')
            except TimeoutExpired:
                process.kill()
                logger.exception(f"Package: Installation failed [Application={application}]")
                raise

            logger.info(f"Application: Successfully installed [Application={application}]")
            # TODO P3: Put list of downloaded packages into cache
        else:
            raise NotImplementedError(f"Application: Install is currently not supported [Package Manager={self.package_manager}]")

    def uninstall_package(self, package):
        logger.info(f"Package: Starting uninstall [Package={package}]")

        if self.package_manager == PackageManager.BREW:
            command = f"brew uninstall {package}"
            command_result = subprocess.run(command.split())

            if command_result.returncode != 0:
                # TODO P3: replace with custom error
                logger.warning(f"Package: Failed to uninstall [Package={package}]")
                raise RuntimeError(f"Package: An issue occurred when trying to uninstall `{package}`")
        elif self.package_manager == PackageManager.APT or self.package_manager == PackageManager.YUM:
            command = f"{self.package_manager} uninstall -y {package}"

            process = Popen(['sudo', '-S'] + command.split(), stdin=PIPE, stderr=PIPE, universal_newlines=True)
            try:
                process.communicate(self.sudo_password + '\n')
            except TimeoutExpired:
                process.kill()
                logger.exception(f"Package: Failed to uninstall [Package={package}]")
                raise
        else:
            raise NotImplementedError(f"Package: Uninstall is currently not supported [Package Manager={self.package_manager}]")

        logger.info(f"Package: Successfully uninstalled [Package={package}]")

    def uninstall_application(self, application):
        logger.info(f"Application: Starting uninstall [Application={application}]")

        if self.package_manager == PackageManager.BREW:
            command = f"brew cask uninstall {application}"
            command_result = subprocess.run(command.split())

            if command_result.returncode != 0:
                logger.info(f"Application: Failed to uninstall [Application={application}]")
                raise RuntimeError(f"Application: An issue occurred when trying to uninstall `{application}`")

            logger.info(f"Application: Successfully uninstalled [Application={application}]")
        else:
            self.uninstall_package(application)


"""
Helper classes
"""


class GitRemoteProgress(RemoteProgress):
    """
    An object passed as a callback that will display
    a progressbar while downloading files from the git
    remote
    """

    def __init__(self):
        super().__init__()
        self.progress_bar = None
        self.is_done = False

    def update(self, op_code, cur_count, max_count=None, message=''):
        if self.is_done:
            return

        if self.progress_bar is None:
            self.progress_bar = ProgressBar(max_value=max_count)
            self.progress_bar.start()

        if cur_count == max_count:
            self.progress_bar.finish()
            self.is_done = True
        else:
            self.progress_bar.update(cur_count)


"""
Parsing: Global pydotfiles configs
"""


def get_pydotfiles_config_data_with_override(override_config_local_directory, override_config_remote_repo, cache_directory):
    # Early return: All overrides provided
    if override_config_local_directory is not None and override_config_remote_repo is not None:
        return override_config_local_directory, override_config_remote_repo

    # Applies config-file overrides
    config_local_directory, config_remote_repo = load_pydotfiles_config_data(cache_directory)

    # Adds in the overrides
    if override_config_local_directory is not None:
        config_local_directory = override_config_local_directory

    if override_config_remote_repo is not None:
        config_remote_repo = override_config_remote_repo

    return config_local_directory, config_remote_repo


def load_pydotfiles_config_data(cache_directory):
    """
    Loads in config-file overrides
    if they exist
    """
    config_local_directory = DEFAULT_PYDOTFILES_CONFIG_LOCAL_DIRECTORY
    config_remote_repo = DEFAULT_CONFIG_REMOTE_REPO

    config_data = cache_directory.read_from_config()

    config_file_local_directory = config_data.get('local_directory')
    if config_file_local_directory is not None:
        config_local_directory = config_file_local_directory

    config_file_remote_repo = config_data.get('remote_repo')
    if config_file_remote_repo is not None:
        config_remote_repo = config_file_remote_repo

    return config_local_directory, config_remote_repo


def write_pydotfiles_config_data(cache_directory, config_repo_local, config_repo_remote):
    cache_directory.write_to_config({
        'local_directory': config_repo_local,
        'remote_repo': config_repo_remote
    })


def load_config_repo_remote(config_repo_local):
    config = configparser.ConfigParser()
    config.read_file(open(f'{config_repo_local}/.git/config', 'r'))
    return config["remote \"origin\""]["url"]


"""
Parsing: Module-level configs
"""


def parse_operating_system_config(os_config, cache_directory, directory):
    """
    Deserializes a single OS dictionary
    to an OS object
    """
    if os_config is None:
        return None
    else:
        name = os_config.get('name')
        shell = os_config.get('shell')
        packages = os_config.get('packages')
        applications = os_config.get('applications')

        raw_default_dock = os_config.get('default_dock')
        default_dock_applications = None if raw_default_dock is None else set(raw_default_dock)

        default_settings_file_name = os_config.get('default_settings_file')

        default_settings = None
        if default_settings_file_name is not None:
            default_settings_file_path = os.path.join(directory, default_settings_file_name)
            default_settings = get_os_default_settings(default_settings_file_path)

        return OperatingSystem(
            name=name,
            shell=shell,
            packages=packages,
            applications=applications,
            cache_directory=cache_directory,
            default_dock=default_dock_applications,
            settings=default_settings,
            dock_manager=DockManager(),
        )


def parse_action_configs(action_configs, directory, symlinks, other_files):
    is_sudo_used = False

    if action_configs is None:
        return [], is_sudo_used

    action_map = {}

    for action_group in action_configs:
        action = FileActionType.from_string(action_group.get('action'))
        file_map = action_group.get('files')

        use_sudo = action_group.get('sudo', False)
        make_hidden = action_group.get('hidden', False)
        use_absolute_path = action_group.get('absolute', False)

        if use_sudo:
            is_sudo_used = True

        action_map.update(deserialize_file_action(action, file_map, use_sudo, make_hidden, use_absolute_path, directory, symlinks, other_files))

    return list(action_map.values()), is_sudo_used


"""
Parsing: FileActions
"""


def deserialize_file_action(action, file_map, use_sudo, make_hidden, use_absolute_path, directory, symlinks, other_files):
    action_map = {}

    action_map.update(deserialize_file_action_expansion(action, file_map, use_sudo, make_hidden, symlinks, other_files))

    for origin, destination in file_map.items():
        absolute_origin = resolve_file_action_absolute_origin(action, origin, directory, use_absolute_path)
        absolute_destination = resolve_file_action_absolute_destination(destination, make_hidden)
        action_map[(action, absolute_destination)] = FileAction(action, absolute_origin, absolute_destination, use_sudo)

    return action_map


def deserialize_file_action_expansion(action, file_map, use_sudo, make_hidden, symlinks, other_files):
    action_map = {}

    # Deals with the case of "all"
    all_case_destination_base = file_map.get('*')
    if all_case_destination_base is not None:
        if action == FileActionType.SYMLINK:
            for origin in symlinks:
                destination_file = f"{all_case_destination_base}/{'.' if make_hidden else ''}{os.path.basename(origin).replace('.symlink', '')}"
                absolute_destination = os.path.expanduser(destination_file)
                action_map[(action, absolute_destination)] = FileAction(action, origin, absolute_destination, use_sudo)
        elif action == FileActionType.COPY:
            for origin in other_files:
                destination_file = f"{all_case_destination_base}/{'.' if make_hidden else ''}{os.path.basename(origin)}"
                absolute_destination = os.path.expanduser(destination_file)
                action_map[(action, absolute_destination)] = FileAction(action, origin, absolute_destination, use_sudo)
        else:
            raise NotImplementedError(f"File Action: The file action `{action}` currently doesn't support expansion, please open an issue on github if you would like to have it")
        file_map.pop('*')

    return action_map


def resolve_file_action_absolute_origin(action, origin, directory, use_absolute_path):
    # Early return
    if use_absolute_path:
        return os.path.expanduser(origin)

    # Gives back a relative origin based off of the action type
    if action == FileActionType.SYMLINK:
        return f"{os.path.join(directory, origin)}.symlink"
    elif action == FileActionType.COPY:
        return f"{os.path.join(directory, origin)}"
    else:
        raise NotImplementedError(f"File Action: The action `{action}` is currently not supported")


def resolve_file_action_absolute_destination(destination, make_hidden):
    if make_hidden:
        stripped_file_name = remove_prefix(f"{os.path.basename(destination)}", '.')
        full_file_name = f"{'.' if make_hidden else ''}{stripped_file_name}"
    else:
        full_file_name = os.path.basename(destination)
    destination_directory = os.path.dirname(destination)
    full_destination = os.path.join(destination_directory, full_file_name)

    return os.path.expanduser(full_destination)


"""
Object creation: Modules
"""


def load_active_modules(config_repo_local, active_modules, host_os, cache_directory):
    modules = {}
    is_sudo_used = False

    module_names = get_module_names(config_repo_local) if active_modules is None else active_modules
    for module_name in module_names:
        settings_file = None

        start_file = None
        post_file = None

        undo_start_file = None
        undo_post_file = None

        module_directory = os.path.join(config_repo_local, module_name)
        module_symlinks = []
        module_generic_files = []

        for module_file in os.listdir(module_directory):
            full_module_file_path = os.path.join(module_directory, module_file)
            if module_file == 'start':
                start_file = full_module_file_path
                continue

            if module_file == 'undo-start':
                undo_start_file = full_module_file_path
                continue

            if module_file == 'settings.yaml' or module_file == 'settings.json':
                settings_file = full_module_file_path
                continue

            if module_file == 'post':
                post_file = full_module_file_path
                continue

            if module_file == 'undo-post':
                undo_post_file = full_module_file_path
                continue

            if module_file.endswith(".symlink"):
                module_symlinks.append(full_module_file_path)
            else:
                module_generic_files.append(full_module_file_path)

        modules[module_name] = Module(
            name=module_name,
            directory=module_directory,
            start_file=start_file,
            post_file=post_file,
            undo_start_file=undo_start_file,
            undo_post_file=undo_post_file,
            settings_file=settings_file,
            symlinks=module_symlinks,
            other_files=module_generic_files,
            host_os=host_os,
            cache_directory=cache_directory
        )

        if modules[module_name].is_sudo_used:
            is_sudo_used = True

    return modules, is_sudo_used


def get_module_names(config_repo_local):
    return [module_name for module_name in os.listdir(config_repo_local) if os.path.isdir(os.path.join(config_repo_local, module_name)) and module_name != ".git"]


"""
Helper functions
"""


def run_script(script_file):
    logger.info(f"Script: Running script [File={script_file}]")
    command_result = subprocess.run(script_file)

    if command_result.returncode == 0:
        logger.info(f"Script: Successfully ran script [File={script_file}]")
    else:
        logger.warning(f"Script: Failed to run script [File={script_file}, stderr={command_result.stderr.decode()}")



import json
import yaml
from pkg_resources import resource_stream


"""
Utility helpers
"""


def load_test_data(version, file_name):
    if file_name.endswith(".json"):
        return json.load(resource_stream(f"tests.resources.validator.{version}", file_name))
    elif file_name.endswith(".yaml") or file_name.endswith(".yml"):
        return yaml.load(resource_stream(f"tests.resources.validator.{version}", file_name))



# General imports
from platform import mac_ver
from logging import getLogger

# Project imports
from pydotfiles.v4.common.alpha import DefaultSettings
from pydotfiles.v4.common.alpha.default_settings import MacVersion
from pydotfiles.v4.common.alpha.os import PackageManager, OperatingSystem

logger = getLogger(__name__)

##
# OS packages installation
##


def template_os_packages(data: OperatingSystem) -> str:
    if len(data.packages) == 0:
        return ""

    package_manager_package_installer = {
        PackageManager.yum: __template_yum_packages,
        PackageManager.brew: __template_homebrew_packages,
        PackageManager.apt: __template_apt_packages,
    }
    return package_manager_package_installer.get(data.package_manager)(data)


def __template_yum_packages(data: OperatingSystem) -> str:
    logger.error("TODO: Implement template yum packages")
    return ""


def __template_homebrew_packages(data: OperatingSystem) -> str:
    homebrew_packages = f"homebrew_packages=({' '.join(data.packages)})\n"

    install_function = r"""
    # Historically before 2019, Homebrew allowed the use of flags when installing packages,
    # such as using `brew install grep --with-default-names`. With Homebrew 2, this has gone
    # away, and we rely instead on gnubin as written here: https://stackoverflow.com/a/60063337

    tmp_homebrew_cache_file=/tmp/pydotfiles-homebrew-cache
    if [[ ! -f /tmp/pydotfiles-homebrew-cache ]]; then
        tmp_homebrew_cache_file=$(mktemp /tmp/pydotfiles-homebrew-cache)
    fi
    
    # chown's the homebrew /usr/local directory since otherwise this issue pops up:
    # https://github.com/Homebrew/legacy-homebrew/issues/44938
    if [[ $(ls -alh /usr/local | grep share) == "" ]]; then
        success "Brew: Ownership of homebrew packages and applications is already owned by the user"
    else
        info "Brew: An issue was detected where the ownership of homebrew packages and applications was set to root instead of $USER. Executing a chown now."        
        if sudo chown -R $USER /usr/local/share ; then
            success "Brew: Ownership of homebrew packages and applications is now set to $USER"
        else
            fail "Brew: Ownership of homebrew packages and applications failed to be fixed"
        fi
    fi

    
    function check_homebrew_package() {
        package_name=$1

        if [[ $(grep "${package_name}" "${tmp_homebrew_cache_file}") == "" ]]; then
            info "Homebrew: Package ${package_name} has not been installed yet, installing now"
            if brew install "${package_name}" ${flagged_packages[${package_name}]} &> /dev/null ; then
                success "Homebrew: Package ${package_name} is now installed"
            else
                fail "Homebrew: Package ${package_name} failed to install"
            fi
        else
            success "Homebrew: Package ${package_name} is already installed"
        fi
    }

    function cache_brew_packages() {
        # -t sorts by time modified
        # -1 makes the packages each go on a single line
        brew list -t1 > "$tmp_homebrew_cache_file"
    }

    cache_brew_packages

    # Updates and upgrades brew packages
    info "Homebrew: Updating brews now"
    if brew update; then
        success  "Homebrew: Brews have been updated"
    else
        fail "Homebrew: Brews failed to be updated"
    fi

    info "Homebrew: Upgrading brews now"
    if brew upgrade; then
        success "Homebrew: Brews have been upgraded"
    else
        fail "Homebrew: Brews failed to upgrade"
    fi

    # Checks and installs any missing packages
    info "Homebrew: Checking installed packages"
    for package in "${homebrew_packages[@]}"; do
        check_homebrew_package "${package}"
    done
    """
    return homebrew_packages + install_function


def __template_apt_packages(data: OperatingSystem) -> str:
    apt_packages = f"apt_packages=({' '.join(data.packages)})\n"

    package_installation_function = r"""
    function check_and_install_package () {
        info "APT: Checking for $1"
        if dpkg -s "$1" > /dev/null 2>&1 ; then
            success "APT: $1 is already installed"
        else
            info "APT: $1 not found, installing now"
            if sudo apt-get install "$1" -y &> /dev/null ; then
                success "APT: $1 successfully installed"
            else
                fail "APT: $1 failed to install"
            fi
        fi
    }
    
    function update_and_upgrade () {
        # Updates & upgrades
        info "APT: Updating packages"
        if sudo apt-get update -y > /dev/null ; then
            success "APT: Packages were updated"
        else
            fail "APT: Packages were unable to be updated"
        fi
    
        info "APT: Upgrading packages"
        if sudo apt-get dist-upgrade -y > /dev/null ; then
            success "APT: Packages were upgraded"
        else
            fail "APT: Packages were unable to be upgraded"
        fi
    }
    
    function auto_remove () {
        # Auto removes any unnecessary packages
        info "APT: Auto removing any unnecessary packages"
        if sudo apt-get autoremove -y > /dev/null ; then
            success "APT: All unnecessary packages removed"
        else
            fail "APT: Unable to remove unnecessary packages"
        fi
    }
    
    # Does first-time initial setup if the user has root/sudo privileges
    if [[ $(groups | grep 'root\|admin\|sudo') != "" ]]; then
        # Does server setup
        update_and_upgrade
        
        for package in "${apt_packages[@]}"; do
            check_and_install_package "${package}"
        done
        auto_remove
    else
        warn "APT: $username does not have the correct permissions to install dependencies"
    fi
    """
    return apt_packages + package_installation_function


##
# OS applications installation
##


def template_os_applications(data: OperatingSystem) -> str:
    if len(data.applications) == 0:
        return ""

    package_manager_package_installer = {
        PackageManager.yum: __template_yum_applications,
        PackageManager.brew: __template_homebrew_applications,
        PackageManager.apt: __template_apt_applications,
    }
    return package_manager_package_installer.get(data.package_manager)(data)


def __template_yum_applications(data: OperatingSystem) -> str:
    logger.error("TODO: Implement template yum applications")
    return ""


def __template_homebrew_applications(data: OperatingSystem) -> str:
    cask_applications = f"cask_packages=({' '.join(data.applications)})\n"

    install_function = r"""
    function check_homebrew_cask_package() {
        cask_name=$1

        if [[ $(grep "${cask_name}" "${tmp_homebrew_cache_file}") == "" ]]; then
            info "Homebrew-Cask: Application ${cask_name} has not been installed yet, installing now"
            if brew install "${cask_name}" --cask &> /dev/null ; then
                success "Homebrew Casks: Application ${cask_name} is now installed"
            else
                fail "Homebrew Casks: Application ${cask_name} failed to install"
            fi
        else
            success "Homebrew Casks: Application ${cask_name} is already installed"
        fi
    }

    # Checks and installs any missing packages
    info "Homebrew Casks: Checking application statuses"
    for package in "${cask_packages[@]}"; do
        check_homebrew_cask_package "${package}"
    done
    """
    return cask_applications + install_function


def __template_apt_applications(data: OperatingSystem) -> str:
    logger.error("TODO: Implement template apt applications")
    return ""

##
# OS default dock
##


def template_os_default_dock(data: OperatingSystem) -> str:
    """
    We only support this on macos for now, not sure if there's a way to programmatically setup the desktop
    for a linux GUI (there probably is, but I'm too lazy to figure it out)
    """
    if len(data.default_dock) == 0:
        return ""

    header = """

    ##
    # Removes all non-essential dock apps and replaces them with our designated ones
    ##

    """

    valid_apps = f"""
    # Apps that you want included in the default dock
    valid_apps=("launchpad" "Notes" "iTunes" "appstore" "systempreferences" "firefox")
    """

    dock_setting_function = r"""
    function check_and_manage_dock_apps() {
        app_name=$1
        if [[ $(defaults read com.apple.Dock persistent-apps | grep "${app_name}") == "" ]]; then
            info "Dock: ${app_name} is not set on the dock, setting now"
            defaults write com.apple.dock persistent-apps -array-add \
                "<dict><key>tile-data</key><dict><key>file-data</key><dict><key>_CFURLString</key><string>/Applications/$1.app</string><key>_CFURLStringType</key><integer>0</integer></dict></dict></dict>"
            success "Dock: ${app_name} is now set on the dock"
        else
            success "Dock: ${app_name} is already set on the dock"
        fi
    }

    """

    repopulate_function = """
    function repopulate_all_dock_apps() {
    """

    for default_dock_item in data.default_dock:
        repopulate_function += f"""check_and_manage_dock_apps "{default_dock_item}"
        """

    repopulate_function += """
        killall Dock
    }
    """

    return header + valid_apps + dock_setting_function + repopulate_function + r"""
    # Does this once instead of doing all over each time
    invert_string="grep -v"

    for app in "${valid_apps[@]}"; do
        invert_string+=" -e \"${app}\""
    done

    function check_and_remove_bad_dock_apps() {
        if [[ $(defaults read com.apple.Dock persistent-apps | grep "bundle-identifier" | eval "$invert_string") != "" ]]; then
            info "Dock: Contains non-default applications, killing off now"
            defaults delete com.apple.Dock persistent-apps
            killall Dock
            success "Dock: All non-default applications are now sanitised"
        else
            success "Dock: All non-default applications are already sanitised"
        fi
        info "Dock: Checking for all default app existences"
        repopulate_all_dock_apps
        success "Dock: All default apps are in place"
    }

    check_and_remove_bad_dock_apps
    """


def template_os_default_setting_files(data: OperatingSystem) -> str:
    if len(data.default_settings_files) == 0:
        return ""

    default_settings = ""
    for default_setting in data.default_settings_data:
        default_settings += template_os_default_setting(default_setting)
    return default_settings


def template_os_default_setting(data: DefaultSettings) -> str:
    if not data.should_run(get_current_mac_version()):
        return f"""
        # Command {data.name} is disabled since it was detected to not be in a valid version range or is manually disabled
        # Description: {data.description}
        """

    info_message = f"""

    ##
    # Setting default settings
    ##

    info "{data.name}"

    """

    command_check = f"""
    {f'if [[ $({data.check_command}) == "{data.expected_check_state}" ]]; then' if data.check_output else ''}
    {f'    success "{data.name}: Successfully completed previously"' if data.check_output else ''}
    {f'else' if data.check_output else ''}

    {f'    if {"sudo " if data.sudo else ""}{data.command} ; then'}
    {f'        success "{data.name}: Successfully executed command"'}    
    {f'        {"" if data.post_command is None else data.post_command}'}
    {f'    else'}
    {f'        warn "{data.name}: Failed to complete"'}
    {f'    fi'}

    {f'fi' if data.check_output else ''}
    """
    return info_message + command_check


def get_current_mac_version():
    """
    This should get ripped out eventually, see https://github.com/JasonYao/pydotfiles/issues/81
    for more context
    """
    current_version = mac_ver()[0]
    if current_version is None:
        return MacVersion.BIG_SUR
    return MacVersion.from_version(current_version)
